@@ -8,9 +8,7 @@ import shutil
 import uuid
 import warnings
 import copy 
-import atexit
 from vipy.util import is_email_address
-from vipy.batch import Batch       
 import hashlib
 import pickle
 import time
@@ -18,8 +16,8 @@ import json
 import dill
 from vipy.show import colorlist
 import matplotlib.pyplot as plt
-import vipy.metrics
 import gc 
+import vipy.metrics
 
 
 class Dataset():
@@ -32,46 +30,54 @@ class Dataset():
     >>>     D = D.map(lambda v: v.frame(0))
     >>> list(D)
 
-    >>> D = vipy.dataset.Dataset([vipy.video.RandomScene(), vipy.video.RandomScene()], id='random_scene')
+    Create dataset and export as a directory of json files 
+
+    >>> D = vipy.dataset.Dataset([vipy.video.RandomScene(), vipy.video.RandomScene()])
     >>> D.tojsondir('/tmp/myjsondir')
+    
+    Create dataset from all json or pkl files recursively discovered in a directory and lazy loaded
+
     >>> D = vipy.dataset.Dataset('/tmp/myjsondir')  # lazy loading
 
+    Create dataset from a list of json or pkl files and lazy loaded
+
+    >>> D = vipy.dataset.Dataset(['/path/to/file1.json', '/path/to/file2.json'])  # lazy loading
+    
     .. notes:: Be warned that using the jsondir constructor will load elements on demand, but there are some methods that require loading the entire dataset into memory, and will happily try to do so
     """
 
-    def __init__(self, datasrc, id=None, abspath=True):
+    def __init__(self, objlist, id=None, abspath=True):
 
         self._saveas_ext = ['pkl', 'json']
-        self._id = id if id is not None else uuid.uuid4().hex
+        self._id = id if id is not None else vipy.util.shortuuid(8)
         self._loader = lambda x: x
         self._istype_strict = True
+        self._lazy_loader = False
+        self._abspath = abspath
 
-        if isinstance(datasrc, str) and (vipy.util.isjsonfile(datasrc) or vipy.util.ispklfile(datasrc)):
-            self._objlist = vipy.util.load(datasrc, abspath=abspath)
-            self._id = vipy.util.filetail(datasrc)
-        elif isinstance(datasrc, str) and os.path.isdir(datasrc):
-            if len(vipy.util.listjson(datasrc)) > 0:
-                self._objlist = vipy.util.listjson(datasrc)
-                self._loader = lambda x,b=abspath: vipy.util.lad(x, abspath=b) if vipy.util.isjsonfile(x) else x
-                self._id = datasrc
-                self._istype_strict = False
-            elif os.path.isdir(datasrc) and len(vipy.util.listpkl(datasrc)) > 0:
-                self._objlist = vipy.util.listpkl(datasrc)
-                self._loader = lambda x,b=abspath: vipy.util.load(x, abspath=b) if vipy.util.ispkl(x) else x
-                self._id = datasrc
-                self._istype_strict = False
-            else:
-                self._objlist = []
+        if isinstance(objlist, str) and (vipy.util.isjsonfile(objlist) or vipy.util.ispklfile(objlist)):
+            self._objlist = vipy.util.load(objlist, abspath=abspath)
+        elif isinstance(objlist, str) and os.path.isdir(objlist):
+            self._objlist = vipy.util.listjson(objlist) + vipy.util.listpkl(objlist)
+            self._loader = lambda x,b=abspath:  vipy.util.load(x, abspath=b) if (vipy.util.ispkl(x) or vipy.util.isjsonfile(x)) else x
+            self._istype_strict = False
+            self._lazy_loader = True
+        elif isinstance(objlist, list) and all([(vipy.util.ispkl(x) or vipy.util.isjsonfile(x)) for x in objlist]):
+            self._objlist = objlist 
+            self._loader = lambda x,b=abspath:  vipy.util.load(x, abspath=b) if (vipy.util.ispkl(x) or vipy.util.isjsonfile(x)) else x            
+            self._istype_strict = False
+            self._lazy_loader = True
         else:
-            self._objlist = datasrc
-        
-        self._objlist = tolist(self._objlist)
-        
+            self._objlist = objlist
+
+        self._objlist = tolist(self._objlist)        
         assert len(self._objlist) > 0, "Empty dataset"
-        try:
-            self[0]
-        except Exception as e:
-            raise ValueError('Dataset loader failed with exception "%s"' % str(e))
+
+        if self._lazy_loader:
+            try:
+                self[0]
+            except Exception as e:
+                raise ValueError('Invalid dataset - Lazy load failed with error "%s"' % str(e))
 
     def __repr__(self):
         return str('<vipy.dataset: id="%s", len=%d, type=%s>' % (self.id(), len(self), str(type(self[0])) if len(self)>0 else 'None'))
@@ -91,6 +97,20 @@ class Dataset():
             
     def __len__(self):
         return len(self._objlist)
+        
+    def json(self, encode=True):
+        r = vipy.util.class_registry()
+        d = {k:v for (k,v) in self.__dict__.items() if not k == '_loader'}
+        d['_objlist'] = [(str(type(v)), v.json(encode=False)) if str(type(v)) in r else v for v in self._objlist]
+        return json.dumps(d) if encode else d
+        
+    @classmethod
+    def from_json(cls, s):
+        r = vipy.util.class_registry()
+        d = json.loads(s) if not isinstance(s, dict) else s  
+        return cls(objlist=[r[x[0]](x[1]) if (isinstance(x, tuple) and x[0] in r) else x for x in d['_objlist']],
+                   id=d['_id'],
+                   abspath=d['_abspath'])                            
 
     def id(self, n=None):
         """Set or return the dataset id"""
@@ -374,20 +394,47 @@ class Dataset():
         objlist = [obj for obj in self if f(obj)]
         return [] if (len(objlist) == 0 or n == 0) else (objlist[0] if n==1 else objlist[0:n])
 
-    def jsondir(self, outdir, verbose=True, rekey=False):
-        """Export all objects to a directory of JSON files"""
+    def jsondir(self, outdir=None, verbose=True, rekey=False, bycategory=False, byfilename=False, abspath=True):
+        """Export all objects to a directory of JSON files.
+    
+           Usage:
+
+           >>> D = vipy.dataset.Dataset(...).jsondir('/path/to/jsondir')
+           >>> D = vipy.util.load('/path/to/jsondir')   # recursively discover and lazy load all json files 
+
+           Args:
+               outdir [str]:  The root directory to store the JSON files
+               verbose [bool]: If True, print the save progress
+               rekey [bool] If False, use the instance ID of the vipy object as the filename for the JSON file, otherwise assign a new UUID_dataset-index
+               bycategory [bool]: If True, use the JSON structure '$OUTDIR/$CATEGORY/$INSTANCEID.json'
+               byfilename [bool]: If True, use the JSON structure '$FILENAME.json' where $FILENAME is the underlying media filename of the vipy object
+               abspath [bool]: If true, store absolute paths to media in JSON.  If false, store relative paths to media from JSON directory
+
+           Returns:
+               outdir: The directory containing the JSON files.
+        """
         assert self._isvipy()
-        print('[vipy.dataset]: exporting %d json files to "%s"...' % (len(self), outdir))
-        vipy.util.remkdir(outdir)  # to avoid race condition
-        for (k,v) in enumerate(self):
-            f = vipy.util.save(v, os.path.join(outdir, ('%s.json' % v.instanceid()) if not rekey else '%s_%d.json' % (uuid.uuid4().hex, k)))
+        assert outdir is not None or byfilename 
+        assert not byfilename and bycategory
+
+        if outdir is not None:
+            vipy.util.remkdir(outdir) 
+        if bycategory:
+            tojsonfile = lambda v,k: os.path.join(outdir, v.category(), ('%s.json' % v.instanceid()) if not rekey else ('%s_%d.json' % (uuid.uuid4().hex, k)))
+        elif byfilename:
+            tojsonfile = lambda v,k: vipy.util.toextension(v.filename(), '.json')
+        else:
+            tojsonfile = lambda v,k: os.path.join(outdir, ('%s.json' % v.instanceid()) if not rekey else '%s_%d.json' % (uuid.uuid4().hex, k))
+        
+        for (k,v) in enumerate(self):            
+            f = vipy.util.save(v.clone().relpath(start=filepath(tojsonfile(v,k))) if not abspath else v.clone().abspath(), tojsonfile(v,k))
             if verbose:
                 print('[vipy.dataset.Dataset][%d/%d]: %s' % (k, len(self), f))
         return outdir
 
-    def tojsondir(self, outdir, rekey=False, verbose=True):
-        """Alias for jsondir"""
-        return self.jsondir(outdir, verbose=verbose, rekey=rekey)
+    def tojsondir(self, outdir=None, verbose=True, rekey=False, bycategory=False, byfilename=False, abspath=True):
+        """Alias for `vipy.dataset.jsondir`"""
+        return self.jsondir(outdir, verbose=verbose, rekey=rekey, bycategory=bycategory, byfilename=byfilename, abspath=abspath)
     
     def takelist(self, n, category=None, canload=False):
         """Take n elements of selected category and return list"""
@@ -401,6 +448,11 @@ class Dataset():
     def take_per_category(self, n, id=None, canload=False):
         return Dataset([v for c in self.categories() for v in self.takelist(n, category=c, canload=canload)], id=id)
     
+    def shuffle(self):
+        """Randomly permute elements in this dataset"""
+        self._objlist.sort(key=lambda x: random.random())  # in place
+        return self
+
     def split(self, trainfraction=0.9, valfraction=0.1, testfraction=0, seed=42):
         """Split the dataset by category by fraction so that video IDs are never in the same set"""
         assert self._isvipy(), "Invalid input"
@@ -432,36 +484,57 @@ class Dataset():
         csv = [v.csv() for v in self.list]        
         return vipy.util.writecsv(csv, csvfile) if csvfile is not None else (csv[0], csv[1:])
 
-    def map(self, f_transform, model=None, dst=None, id=None, checkpoint=False, strict=False, ascompleted=True, chunks=128):        
+    def map(self, f_map, model=None, dst=None, id=None, strict=False, ascompleted=True, chunks=128, ordered=False):        
         """Distributed map.
 
         To perform this in parallel across four processes:
 
+        >>> D = vipy.dataset.Dataset(...)
         >>> with vipy.globals.parallel(4):
-        >>>     self.map(lambda v: ...)
+        >>>     D.map(lambda v: ...)
 
         Args:
-            f_transform: [lambda] The lambda function to apply in parallel to all elements in the dataset
+            f_map: [lambda] The lambda function to apply in parallel to all elements in the dataset.  This must return a JSON serializable object
             model: [torch.nn.Module] The model to scatter to all workers
             dst: [str] The ID to give to the resulting dataset
             id: [str] The ID to give to the resulting dataset (parameter alias for dst)
-            checkpoint: [bool] If trye, checkpoint the map operation
             strict: [bool] If true, raise exception on map failures, otherwise the map will return None for failed elements
             ascompleted: [bool] If true, return elements as they complete
+            ordered: [bool] If true, preserve the order of objects in dataset as returned from distributed processing
 
         Returns:
-            A `vipy.dataset.Dataset` containing the elements f_transform(v).  This operation is order preserving.
+            A `vipy.dataset.Dataset` containing the elements f_map(v).  This operation is order preserving.
 
+        .. note:: 
+            - This dataset must contain vipy objects of types defined in `vipy.util.class_registry` or JSON serializable objects
+            - Serialization of large datasets can take a while, kick it off to a distributed dask scheduler and go get lunch
+            - This method uses dask distributed and `vipy.batch.Batch` operations
+            - All vipy objects are JSON serialized prior to parallel map to avoid reference cycle garbage collection which can introduce instabilities
+            - Due to chunking, all error handling is caught by this method.  Use `vipy.batch.Batch` to leverage dask distributed futures error handling.
+            - Operations must be chunked and serialized because each dask task comes with overhead, and lots of small tasks violates best practices
+            - Serialized results are deserialized by the client and returned a a new dataset
         """
-        assert callable(f_transform)
+        assert callable(f_map)
+        from vipy.batch import Batch   # requires pip install vipy[all]
 
-        # Operations must be chunked because each dask task comes with overhead, and lots of small tasks violates best practices
-        # - Each chunk is sent to a worker and processed in parallel
-        # - Chunks are scattered to workers 
-        B = Batch(vipy.util.chunklist(self._objlist, chunks), strict=strict, as_completed=ascompleted, checkpoint=checkpoint, warnme=False, minscatter=chunks)
-        f = lambda x, f_loader=self._loader: f_transform(f_loader(x))  # for closure capture
-        V = B.map(lambda X,f=f: [f(x) for x in X]).result() if not model else B.scattermap(lambda X, f=f: [f(x) for x in X], model).result()
-        return Dataset(V, id=dst if dst is not None else id).flatten()
+        # Distributed map using vipy.batch
+        f_serialize = lambda v,d=vipy.util.class_registry(): (str(type(v)), v.json()) if str(type(v)) in d else (None, json.dumps(v, ensure_ascii=False))  # fallback on JSON dumps/loads
+        f_deserialize = lambda x,d=vipy.util.class_registry(): d[x[0]](x[1])  # with closure capture
+        f_catcher = vipy.util.catcher  # catch exceptions when executing lambda and return (True, result) or (False, exception)
+        S = [f_serialize(v) for v in self._objlist]  # local serialization
+        B = Batch(vipy.util.chunklist(S, chunks), strict=strict, as_completed=ascompleted, warnme=False, minscatter=chunks, ordered=ordered)        
+        if model is None:
+            f = lambda x, f_loader=self._loader, f_serializer=f_serialize, f_deserializer=f_deserialize, f_map=f_map, f_catcher=f_catcher: f_serializer(f_catcher(f_map, f_loader(f_deserializer(x))))  # with closure capture
+            S = B.map(lambda X,f=f: [f(x) for x in X]).result()  # chunked, with caught exceptions
+        else:
+            f = lambda net, x, f_loader=self._loader, f_serializer=f_serialize, f_deserializer=f_deserialize, f_map=f_map, f_catcher=f_catcher: f_serializer(f_catcher(f_map, net, f_loader(f_deserializer(x))))  # with closure capture
+            S = B.scattermap((lambda net, X, f=f: [f(net, x) for x in X]), model).result()  # chunked, scattered, caught exceptions
+        V = [f_deserialize(x) for s in S for x in s]  # Local deserialization and chunk flattening
+        (good, bad) = ([r for (b,r) in V if b], [r for (b,r) in V if not b])  # catcher returns (True, result) or (False, exception string)
+        if len(bad) > 0:
+            print('[vipy.dataset.Dataset.map]: Exceptions in map distributed processing:\n%s' % (len(bad), len(self), str(bad)))
+            print('[vipy.dataset.Dataset.map]: %d/%d items failed' % (len(bad), len(self)))
+        return Dataset(good, id=dst if dst is not None else id)
 
     def localmap(self, f):
         self._objlist = [f(v) for v in self]
@@ -486,6 +559,7 @@ class Dataset():
         return len([v for v in self if f is None or f(v)])
 
     def countby(self, f=lambda v: v.category()):
+        """Count the number of elements that return the same value from the lambda function"""
         assert self._isvipy()
         assert f is None or callable(f)
         return vipy.util.countby(self, f)
@@ -501,7 +575,6 @@ class Dataset():
         
         d = self.countby(lambda v: v.category())
         if outfile is not None:
-            from vipy.metrics import histogram            
             (categories, freq) = zip(*reversed(sorted(list(d.items()), key=lambda x: x[1])))  # decreasing frequency
             barcolors = ['blue' if category_to_barcolor is None else category_to_barcolor[c] for c in categories]
             xlabels = [f_category_to_xlabel(c) for c in categories]
@@ -551,24 +624,21 @@ class Dataset():
         assert self._isvipy()
         d = {k:np.mean([v[1] for v in v]) for (k,v) in groupbyasdict([(a.category(), len(a)) for v in self.list() for a in v.activitylist()], lambda x: x[0]).items()}
         if outfile is not None:
-            from vipy.metrics import histogram
-            histogram(d.values(), d.keys(), outfile=outfile, ylabel='Duration (frames)', fontsize=6)            
+            vipy.metrics.histogram(d.values(), d.keys(), outfile=outfile, ylabel='Duration (frames)', fontsize=6)            
         return d
 
     def duration_in_seconds(self, outfile=None, fontsize=6):
         assert self._isvipy()
         d = {k:np.mean([v[1] for v in v]) for (k,v) in groupbyasdict([(a.category(), len(a)/v.framerate()) for v in self.list() for a in v.activitylist()], lambda x: x[0]).items()}
         if outfile is not None:
-            from vipy.metrics import histogram
-            histogram(d.values(), d.keys(), outfile=outfile, ylabel='Duration (seconds)', fontsize=fontsize)            
+            vipy.metrics.histogram(d.values(), d.keys(), outfile=outfile, ylabel='Duration (seconds)', fontsize=fontsize)            
         return d
 
     def framerate(self, outfile=None):
         assert self._isvipy()
         d = vipy.util.countby([int(round(v.framerate())) for v in self.list()], lambda x: x)
         if outfile is not None:
-            from vipy.metrics import pie
-            pie(d.values(), ['%d fps' % k for k in d.keys()], explode=None, outfile=outfile,  shadow=False)
+            vipy.metrics.pie(d.values(), ['%d fps' % k for k in d.keys()], explode=None, outfile=outfile,  shadow=False)
         return d
         
         
@@ -578,8 +648,7 @@ class Dataset():
         d = [len(v) if (max is None or len(v)<= max) else max for (k,v) in groupbyasdict(self.list(), lambda v: v.videoid()).items()]
         d = {k:v for (k,v) in sorted(vipy.util.countby(d, lambda x: x).items(), key=lambda x: x[1], reverse=True)}
         if outfile is not None:
-            from vipy.metrics import histogram
-            histogram(d.values(), d.keys(), outfile=outfile, ylabel='Frequency', xlabel='Activities per video', fontsize=6, xrot=None)            
+            vipy.metrics.histogram(d.values(), d.keys(), outfile=outfile, ylabel='Frequency', xlabel='Activities per video', fontsize=6, xrot=None)            
         return d
 
     def boxsize(self, outfile=None, category_to_color=None, categories=None):
@@ -673,7 +742,9 @@ class Dataset():
         This is useful for fast loading of datasets that contain many videos.
 
         """
-        import vipy.torch
+        import vipy.torch    # lazy import, requires vipy[all] 
+        from vipy.batch import Batch   # requires pip install vipy[all]
+
         assert self._is_vipy_scene()
         outdir = vipy.util.remkdir(outdir)
         vipy.batch.Batch(self.list(), as_completed=True).map(lambda v, f=f_video_to_tensor, outdir=outdir, n_augmentations=n_augmentations: vipy.util.bz2pkl(os.path.join(outdir, '%s.pkl.bz2' % v.instanceid()), [f(v.print(sleep=sleep).clone()) for k in range(0, n_augmentations)]))
