@@ -224,20 +224,63 @@ class Video(vipy.video.Video):
 class Flow(object):
     """vipy.flow.Flow() class"""
     
-    def __init__(self, flowiter=10, flowdim=None):
+    def __init__(self, flowiter=10, flowdim=256, gpu=None):
         self._mindim = flowdim
         self._levels = 3
         self._winsize = 7
         self._poly_n = 5
         self._poly_sigma = 1.2
         self._flowiter = flowiter
-        
+        self._gpu = gpu
+        if gpu == True:
+            try:
+                f = cv2.cuda_FarnebackOpticalFlow
+            except:
+                warnings.warn('OpenCV not CUDA enabled - GPU acceleration is unavailable')
+                self._gpu = None
+
     def __call__(self, im, imprev=None, flowstep=1, framestep=1):
         return self.videoflow(im, flowstep, framestep) if imprev is None else self.imageflow(im, imprev)
 
-    def _numpyflow(self, img, imgprev):
+    def _numpyflow_gpu(self, img, imgprev):
+        """Optical flow on GPU"""
+
+        # To compile CUDA enabled opencv (YUCK):
+        #
+        #   sh> python3 -m venv /path/to/myvirtualenv
+        #   sh> source /path/to/myvirtualenv/bin/activate   # for python bindings
+        #   sh> git clone --recursive https://github.com/opencv/opencv-python.git
+        #   sh> cd opencv-python
+        #   sh> export ENABLE_CONTRIB=1 ENABLE_HEADLESS=1  # opencv packages
+        #   sh> export CMAKE_ARGS="-DWITH_CUDNN=OFF -DOPENCV_DNN_CUDA=OFF -DWITH_CUDA=ON -DCUDA_ARCH_BIN=5.2 -DCMAKE_CXX_STANDARD=11 -DPYTHON3_EXECUTABLE=$(which python) -DINSTALL_PYTHON_EXAMPLES=OFF -DCMAKE_CXX_STANDARD_REQUIRED=ON -DCMAKE_CXX_FLAGS=\"-std=c++11\""
+        #   sh> pip wheel . --verbose
+        #   sh> pip install /path/to/opencv_contrib_python_headless-4.5.3+c1cc7e5-cp36-cp36m-linux_x86_64.whl   # path output at compile time
+        #   sh> cd ..  # do not import from within-source
+        #   sh> python
+        #   >>> import cv2
+        #   >>> cv2.cuda_FarnebackOpticalFlow
+        #
+        # See also: 
+        #   https://github.com/opencv/opencv-python#manual-builds
+        #   https://learnopencv.com/getting-started-opencv-cuda-module/
+        #   https://developer.nvidia.com/blog/opencv-optical-flow-algorithms-with-nvidia-turing-gpus/
+
+        (gpu_img, gpu_imgprev) = (cv2.cuda_GpuMat(), cv2.cuda_GpuMat())
+        gpu_img.upload(img)
+        gpu_imgprev.upload(imgprev)
+        gpu_flow = cv2.cuda_FarnebackOpticalFlow.create(self._levels, 0.5, False, self._winsize, self._flowiter, self._poly_n, self._poly_sigma, cv2.OPTFLOW_FARNEBACK_GAUSSIAN)
+        gpu_flow = cv2.cuda_FarnebackOpticalFlow.calc(gpu_flow, gpu_img, gpu_imgprev, None)
+        flow = gpu_flow.download()        
+        return vipy.flow.Image(flow)
+
+    def _numpyflow_cpu(self, img, imgprev):
         """Overload this method for custom flow classes"""        
         return Image(cv2.calcOpticalFlowFarneback(img, imgprev, None, 0.5, self._levels, self._winsize, self._flowiter, self._poly_n, self._poly_sigma, cv2.OPTFLOW_FARNEBACK_GAUSSIAN))
+
+    def _numpyflow(self, img, imgprev):
+        """Overload this method for custom flow classes"""        
+        f = self._numpyflow_cpu if self._gpu is None or self._gpu == False else self._numpyflow_gpu
+        return f(img, imgprev)
         
     def imageflow(self, im, imprev):
         """Default opencv dense flow, from im to imprev.  This should be overloaded"""        
@@ -254,11 +297,10 @@ class Flow(object):
         imf = [self.imageflow(v[k], v[max(0, k-flowstep) if keyframe is None else keyframe]) for k in range(0, len(v.load())+framestep, framestep) if k < len(v.load())]
         return Video(np.stack([im.flow() for im in imf]), flowstep, framestep)  # flow only, no objects
 
-    def videoflowframe(self, v, frame, flowstep=1, framestep=1, keyframe=None):
+    def videoflowframe(self, v, frame, duration, flowstep=1, framestep=1, keyframe=None):
         """Computer the videoflow for a single frame"""
         assert isinstance(v, vipy.video.Video)
         assert flowstep == 1 and framestep == 1
-        duration = v.duration_in_frames_of_videofile()
         imf = [self.imageflow(v.frame(k), v.frame(max(0, k-flowstep) if keyframe is None else keyframe)) for k in range(frame, frame+framestep, framestep) if k < duration]
         return imf[0]
 
@@ -270,10 +312,10 @@ class Flow(object):
                for k in range(0, len(v.load()))]
         return Video(np.stack([im.flow() for im in imf]), flowstep=1, framestep=1)  # flow only, no objects
 
-    def keyflowframe(self, v, frame, keystep=None):
+    def keyflowframe(self, v, frame, duration, keystep=None):
         """Compute the keyflow for a single frame"""
         assert isinstance(v, vipy.video.Video)
-        len_v = v.duration_in_frames_of_videofile()
+        len_v = duration
         assert frame < len_v
         (ima, imb, imc) = (v.frame(min(len_v-1, int(keystep*np.round(frame/keystep)))), v.frame(max(0, frame-1)), v.frame(frame))
         return self.imageflow(ima, imb) - self.imageflow(ima, imc)
@@ -290,60 +332,77 @@ class Flow(object):
         """Return a flow field of size (height=H, width=W) consistent with an Euclidean transform parameterized by a 2x2 Rotation and 2x1 translation"""  
         return self.affineflow(np.array([[R[0,0], R[0,1], t[0]], [R[1,0], R[1,1], t[1]]]), H, W)
     
-    def _correspondence(self, imflow, im, border=0.1, contrast=(16.0/255.0), dilate=1.0, validmask=None, maxflow=None):
+    def _correspondence(self, imflow, im, border=0.1, contrast=(16.0/255.0), dilate=1.0, validmask=None, maxflow=None, subsample=1):
         (H,W) = (imflow.height(), imflow.width())
         m = im.clone().dilate(dilate).rectangular_mask() if (dilate  is not None and isinstance(im, vipy.image.Scene) and len(im.objects())>0) else 0  # ignore foreground regions
         b = im.border_mask(int(border*min(W,H))) if border is not None else 0  # ignore borders
         w = np.uint8(np.sum(np.abs(np.gradient(im.clone().greyscale().numpy())), axis=0) < contrast) if contrast is not None else 0  # ignore low contrast regions
         v = (1-np.float32(validmask)) if validmask is not None else 0  # ignore non-valid regions
         x = np.float32(imflow.magnitude() > maxflow) if maxflow is not None else 0  # ignore maxflow region
-        bk = np.nonzero((m+b+w+v+x) == 0)  # indexes for valid flow regions
+        vf = (m+b+w+v+x)  # valid flow regions (non-zero elements)
+        if subsample != 1:
+            assert isinstance(subsample, int) and subsample > 1
+            vf[::subsample, ::subsample] = 0  # zero out (ignore) neighboring flow
+        bk = np.nonzero(vf == 0)  # indexes for valid flow regions
         (X, Y) = np.meshgrid(np.arange(0, im.width()), np.arange(0, im.height()))        
         (fx, fy) = (imflow.dx()[bk].flatten(), imflow.dy()[bk].flatten())  # flow
         (x1, y1) = (X[bk].flatten(), Y[bk].flatten())  # image coordinates
         (x2, y2) = (x1 + fx, y1 + fy)  # destination coordinates
         return (np.stack((x1,y1)), np.stack((x2,y2)))
         
-    def stabilize(self, v, keystep=20, padheightfrac=0.125, padwidthfrac=0.25, padheightpx=None, padwidthpx=None, border=0.1, dilate=1.0, contrast=16.0/255.0, rigid=False, affine=True, verbose=True, strict=True, residual=False, maxflow=None, outfile=None): 
+    def stabilize(self, v, keystep=20, padheightfrac=0.125, padwidthfrac=0.25, padheightpx=None, padwidthpx=None, border=0.1, dilate=1.0, contrast=16.0/255.0, rigid=False, affine=True, verbose=True, strict=True, residual=False, maxflow=None, outfile=None, preload=True): 
         """Affine stabilization to frame zero using multi-scale optical flow correspondence with foreground object keepouts.  
 
-           * v [vipy.video.Scene]:  The input video to stabilize, should be resized to mindim=256
-           * keystep [int]:  The local stabilization step between keyframes (should be <= 30)
-           * padheightfrac [float]:  The height padding (relative to video height) to be applied to output video to allow for vertical stabilization
-           * padwidthfrac [float]:  The width padding (relative to video width) to be applied to output video to allow for horizontal stabilization
-           * padheightpx [int]:  The height padding to be applied to output video to allow for vertical stabilization.  Overrides padheight.
-           * padwidthpx [int]:  The width padding to be applied to output video to allow for horizontal stabilization.  Overrides padwidth.
-           * border [float]:  The border keepout fraction to ignore during flow correspondence.  This should be proportional to the maximum frame to frame flow
-           * dilate [float]:  The dilation to apply to the foreground object boxes to define a foregroun keepout for flow computation
-           * contrast [float]:  The minimum gradient necessary for flow correspondence, to avoid flow on low contrast regions
-           * rigid [bool]:  Euclidean stabilization
-           * affine [bool]:  Affine stabilization
-           * verbose [bool]:  This takes a while to run ...
-           * strict [bool]:  If true, throw an exception on error, otherwise return the original video and set v.hasattribute('unstabilized'), useful for large scale stabilization
-           * outfile [str]: the file path to the stabilized output video
+        Recommended usage:
 
-           * NOTE: The remaining distortion after stabilization is due to: rolling shutter distortion, perspective distortion and non-keepout moving objects in background
-           * NOTE: If the video contains objects, the object boxes will be transformed along with the stabilization 
-           * NOTE: This requires loading videos entirely into memory.  Be careful with stabilizing long videos.
+        >>> v = vipy.video.Scene(filename='/path/to/my/video.mp4').stabilize()
 
+        Args:
+            v: [`vipy.video.Scene`]:  The input video to stabilize, should be resized to mindim=256
+            keystep: [int]  The local stabilization step between keyframes (should be <= 30)
+            padheightfrac: [float] The height padding (relative to video height) to be applied to output video to allow for vertical stabilization
+            padwidthfrac: [float]  The width padding (relative to video width) to be applied to output video to allow for horizontal stabilization
+            padheightpx: [int]  The height padding to be applied to output video to allow for vertical stabilization.  Overrides padheight.
+            padwidthpx: [int]  The width padding to be applied to output video to allow for horizontal stabilization.  Overrides padwidth.
+            border: [float]  The border keepout fraction to ignore during flow correspondence.  This should be proportional to the maximum frame to frame flow
+            dilate: [float]  The dilation to apply to the foreground object boxes to define a foregroun keepout for flow computation
+            contrast: [float]  The minimum gradient necessary for flow correspondence, to avoid flow on low contrast regions
+            rigid: [bool]  Euclidean stabilization
+            affine: [bool]  Affine stabilization
+            verbose: [bool]  This takes a while to run so show some progress ...
+            strict: [bool]  If true, throw an exception on error, otherwise return the original video and set v.hasattribute('unstabilized'), useful for large scale stabilization
+            outfile: [str] the file path to the stabilized output video
+            preload [bool]: If true, load the input video into memory before stabilizing.  Fasterm but requires video to fit into memory.
+
+        Returns:
+            A cloned `vipy.video.Scene` with filename=outfile, such that pixels and tracks are background stabilized.
+
+        .. notes::
+            - The remaining distortion after stabilization is due to: rolling shutter distortion, perspective distortion and non-keepout moving objects in background
+            - If the video contains objects, the object boxes will be transformed along with the stabilization 
+            - This requires loading videos entirely into memory.  Be careful with stabilizing long videos.
+            - The returned video has the attribute 'stabilize' which contains the mean and median residual of the flow field relative to the motion model. This can be used for stabilization quality filtering.
+    
         """
         vc = v.clone()  # clone to avoid memory leaks in distributed processing
+        vc = vc.saveas(tempMP4()) if vc.isloaded() else vc  # dump to temp file if loaded
         
         assert isinstance(vc, vipy.video.Scene), "Invalid input - Must be vipy.video.Scene() with foreground object keepouts for background stabilization"
-        if verbose and min(vc.shape()) > 256:  # shape triggers fine frame fetch
-            warnings.warn('Large video frame size detected - This will take a while ...')
+        if verbose and vc.framerate() != 5:  # stabilization optimized for framerate=5.0Hz
+            warnings.warn('Optimal framerate for vipy.flow.stabilize is 5.0 Hz.  Set this with v.framerate(5).')
+            if vc.framerate() > 5:
+                warnings.warn('High framerate for stabilization - this will take a while ...')
                     
         # Prepare videos
         vv = vc.cropeven()  # make even for zero pad
         (padwidth, padheight) = (int(vv.width()*padwidthfrac) if padwidthpx is None else padwidthpx, int(vv.height()*padheightfrac) if padheightpx is None else padheightpx)  # width() height() triggers single frame fetch
         outfile = premkdir(outfile if outfile is not None else tempMP4())
-        #vs = vv.clone(flush=True).zeropad(padwidth, padheight).load().nofilename().nourl()   # triggers load, this pre-allocates stabilized video buffer to copy into    
-        vs = vv.clone(flush=True, flushfilter=True).filename(outfile if outfile is not None else tempMP4()).nourl().cleartracks()   # does not trigger load
+        vs = vv.clone(flushforward=True, flushfilter=True).filename(outfile if outfile is not None else tempMP4()).nourl().cleartracks()   # does not trigger load
         s = vv.mindim() / float(self._mindim)  # for upsample
-        vvd = vv.clone().mindim(self._mindim)  # downsampled for flow correspondence
-        if vc.duration_in_frames_of_videofile() < keystep:
-            print('[vipy.flow.stabilize]: ERROR - video not long enough for stabilization, returning original video "%s"' % str(v))
-            return vc.setattribute('unstabilized')
+        vvd = vv.clone().mindim(self._mindim)  # downsampled for flow correspondence        
+        if preload:
+            (vv, vvd) = (vv.load(), vvd.load())  # Faster, but requires lots of memory, replace with stream, replace all frame() calls
+        assert preload, "preload=True is required for now"
 
         # Stabilization
         assert rigid is True or affine is True, "Projective stabilization is disabled"
@@ -356,7 +415,10 @@ class Flow(object):
         f_transform_coarse = (lambda A: A[0:2,:])
         f_transform_fine = (lambda A: A[0:2,:]) if (rigid or affine) else (lambda A: A)
         imstabilized = vv.preview(0).rgb().zeropad(padwidth, padheight)  # single frame fetch
-        duration = vv.duration_in_frames_of_videofile()
+        duration = vv.duration_in_frames_of_videofile() if not preload else len(vv)  # FIXME: this may be off if the video filter chain has a different framerate
+        if duration < keystep:
+            print('[vipy.flow.stabilize]: ERROR - video not long enough for stabilization, returning original video "%s"' % str(v))
+            return vc.setattribute('unstabilized')
         r_coarse = []
         frames = []                        
 
@@ -366,14 +428,10 @@ class Flow(object):
                     print('[vipy.flow.stabilize]: %s coarse to fine stabilization ...' % ('Euclidean' if rigid else 'Affine' if affine else 'Projective'))                
 
                 # Optical flow (3x): use downsampled video, do not precompute to save on memory, requires random access to downsampled video
-                try:
-                    im = vv.frame(k)                    
-                    imf = self.videoflowframe(vvd, k, framestep=1, flowstep=1)
-                    imfk1 = self.keyflowframe(vvd, k, keystep=keystep)
-                    imfk2 = self.keyflowframe(vvd, k, keystep=duration//2)
-                except:
-                    print('[vipy.flow.stabilize]: graceful early exit')
-                    break  # FIXME: duration is sometimes off by one causing frame(k) to error, early exit here
+                im = vv.frame(k)                    
+                imf = self.videoflowframe(vvd, k, duration=duration, framestep=1, flowstep=1)
+                imfk1 = self.keyflowframe(vvd, k, duration=duration, keystep=keystep)
+                imfk2 = self.keyflowframe(vvd, k, duration=duration, keystep=duration//2)
             
                 # Coarse alignment 
                 imd = im.clone().rescale(1.0 / s)  # downsample
@@ -382,7 +440,7 @@ class Flow(object):
                 (xy_src_k2, xy_dst_k2) = self._correspondence(imfk2, imd, border=border, dilate=dilate, contrast=contrast, maxflow=maxflow)
                 (xy_src, xy_dst) = (np.hstack( (xy_src_k0, xy_src_k1, xy_src_k2) ).transpose(), np.hstack( (xy_dst_k0, xy_dst_k1, xy_dst_k2) ).transpose())  # Nx3
                 try:            
-                    M = f_estimate_coarse(s*xy_src, s*xy_dst, method=cv2.RANSAC, confidence=0.99999, ransacReprojThreshold=0.1, refineIters=16, maxIters=3000)                   
+                    M = f_estimate_coarse(s*xy_src, s*xy_dst, method=cv2.RANSAC, confidence=0.99999, ransacReprojThreshold=0.1, refineIters=16, maxIters=3000)   # upsampled correspondences
                     r_coarse.append(np.mean(np.sqrt(np.sum(np.square(M.dot(homogenize(xy_src[::8].transpose())) - homogenize(xy_dst[::8].transpose())), axis=0))) if (residual and len(xy_src)>8) else 0)
                 except Exception as e:
                     if not strict:
@@ -414,7 +472,7 @@ class Flow(object):
                         return vc.setattribute('unstabilized')  # for provenance
                     else:
                         raise ValueError('[vipy.flow.stabilize]: ERROR - object alignment returned degenerate bounding box for video "%s"' % str(v))                    
-                vss.write( im.array(imstabilized.array()) )
+                vss.write( im.array(imstabilized.array()) )  # assign detections to tracks in stabilized video (vs) output
 
         vs = vs.setattribute('stabilize', {'mean residual':float(np.mean(r_coarse)), 'median residual':float(np.median(r_coarse))}) if residual else vs
         return vs
